/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/syslog.h>
#include <netdb.h>
#include "dixie.h"
#include <quipu/attrvalue.h>
#include <quipu/ds_error.h>

extern int	dosyslog;
extern int	debug;
extern int	bound;
extern int	stream_connected;

struct version {
    long	version_no;	/* version number, in form 0xAABBCCDD, where
				   version is AA.BB.(Alpha,Beta,...)CC and
				   the Alpha, Beta, etc is determined by DD */
    char	version_reqd;	/* 1 if required upgrade, 0 if not */
    char	*version_date;	/* release date in form DD/MM/YY */
    char	*version_msg;	/* text message */
};

#define CLIENT_VERSIONS	4
static struct version client_versions[] = {
    {	0x01000002,		/* 1.0Beta0 */
	0,
	"2/15/91",
	"Initial X.500 Access release"
    },
    {	0x01000102,		/* 1.0Beta1 */
	0,
	"5/03/91",
	"supports photos, bind, and improved error messages"
    },
    {   0x01000003,             /* 1.0 release */
	0,
	"6/21/91",
	"New interface, Misc. bug fixes, Supports Cut/Copy/Paste"
    },
    {   0x01010003,             /* 1.1 release */
	1,
	"10/9/91",
	"Improved network performance, Misc. bug fixes and interface changes"
    },
};


pingop( bp, bplen, rp, rplenp, fromaddr )
    char		*bp;
    int			bplen;
    char		*rp;
    int			*rplenp;
    struct sockaddr_in	fromaddr;
{
    struct hostent	*hp;
    short		code;
    long		version, along;
    int			i, j, unknown;
    char		*p;
    int			err;

    bcopy( bp, (char *)&code, sizeof( short ));
    code = ntohs( code );
    if ( debug ) printf( "ping code is %hd\n", code );

    hp = gethostbyaddr((char *) &(fromaddr.sin_addr.s_addr),
	    sizeof(fromaddr.sin_addr.s_addr), AF_INET);
    if (debug ) printf( "ping code %hd from %s (%s)\n", code,
	    ( hp == NULL ? "unknown" : hp->h_name ),
	    inet_ntoa(fromaddr.sin_addr));
    if ( dosyslog ) {
	syslog( LOG_INFO|LOG_LOCAL1, "ping code %hd from %s (%s)", code,
		( hp == NULL ? "unknown" : hp->h_name ),
		inet_ntoa(fromaddr.sin_addr));
    }

    if ( bplen - sizeof( short ) >= sizeof( long )) {
	/*
	 * we got a version number, check it
	 */
	bcopy( bp + sizeof( short ), &version, sizeof( long ));
	version = ntohl( version );
	for ( i = 0; i < CLIENT_VERSIONS; ++i ) {
	    if ( client_versions[ i ].version_no == version ) break;
	}

	unknown = 0;
	if ( i != CLIENT_VERSIONS - 1 ) {
	    if ( i == CLIENT_VERSIONS ) {
		--i;
		++unknown;
	    } else {
		++i;
	    }
	    p = rp;
	    code = CLIENT_VERSIONS - i;
	    code = htons( code );
	    bcopy( &code, p, sizeof( short ));
	    p += sizeof( short );
	    for ( j = CLIENT_VERSIONS - 1; j >= i; --j ) {
		along = client_versions[ j ].version_no;
		along = htonl( along );
		bcopy( &along, p, sizeof( long ));
		p += sizeof( long );
		if ( unknown ) {
		    *p++ = 2;
		} else {
		    *p++ = client_versions[ j ].version_reqd;
		}
		bcopy( client_versions[ j ].version_date, p,
			strlen( client_versions[ j ].version_date ) + 1 );
		p += strlen( client_versions[ j ].version_date ) + 1; 
		bcopy( client_versions[ j ].version_msg, p,
			strlen( client_versions[ j ].version_msg ) + 1 );
		p += strlen( client_versions[ j ].version_msg ) + 1; 
	    }
	    *rplenp = p - rp - 1;
	    if ( debug ) {
		printf( "version 0x%x current is 0x%x\n", version,
			client_versions[ CLIENT_VERSIONS-1 ].version_no );
	    }
	    if ( dosyslog ) {
		syslog( LOG_INFO|LOG_LOCAL1, "version 0x%x current is 0x%x",
			version, client_versions[ CLIENT_VERSIONS-1 ].version_no );
	    }

	} else {
	    /*
	     * version is current
	     */
	    if ( debug ) {
		printf( "version 0x%x ok\n", version );
	    }
	    if ( dosyslog ) {
		syslog( LOG_INFO|LOG_LOCAL1, "version 0x%x ok", version );
	    }
	    code = htons( 0 );
	    bcopy( &code, rp, sizeof( short ));
	    *rplenp = 2;
	}

    }

    if ( bound )
	read_entry( NULLDN, NULLATTR, 0, 0, &bound, &err, 0 );

    if ( bound == 0 )
	nullbind();

    if ( bound ) {
	return( DXI_SUCCESS );
    } else {
	return( DXI_DSA_DOWN );
    }
}
