/*
 * Copyright (c) 1991, 1992 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#ifndef __STDC__
#include <memory.h>
#endif
#include <dixie.h>
#include "ud.h"

char base[BUFSIZ];		/* search base */
extern int verbose;		/* verbose mode flag */
static int num_picked = 0;	/* used when user picks entry at More prompt */
char answer[MAX_PKT_SIZE];	/* answer returned by DIXIE lib */

dxi_vrfy(dn)
char *dn;
{
	static char *attrs[2] = { OBJECT_CLASS };

	/* verify that this DN exists in the directory */
	(void) dxi_read(dn, attrs, answer);
	if ((dxi_errno == DXI_BADDN) || (dxi_errno == DXI_NAME_ERROR))
		return(0);
	else if (dxi_errno == DXI_SUCCESS)
		return(1);
	else {
		dxi_perror("  Unexpected value of dxi_errno");
		return(0);
	}
}
	

char * find(who)
char *who;
{
	register char *cp;		/* general char pointer */
	int matches;			/* matches found by dxi_search() */
	int choice;			/* entry that user chooses */
	int spaces = 0;			/* if spaces in the name, not # */
	int surname_failed = 0;		/* if a lookup on the surname failed */
	int try_fuzzy = 0;		/* try a fuzzy match */
	static char response[RESP_SIZE];	/* results from user */
	static char *namelist[MAX_NUM_NAMES];	/* names found */
	static char filter[BUFSIZ];		/* search filter */

	/* read attributes */
	static char *read_attrs[] = { COMMON_NAME, TITLE, 
		POSTAL_ADDRESS, PHONE_NUMBER, E_MAIL_ADDRESS, 
		HOME_PHONE, HOME_ADDRESS, OBJECT_CLASS, 
		MEMBERS, DESCRIPTION, OWNER, 
		MLDESCRIPTION, FAX, UNIQNAME,
		NULL };

	/* search attributes */
	static char *attrs[] = { COMMON_NAME, TITLE, OBJECT_CLASS, NULL };

	/* initialize answer buffer */
	(void) memset(answer, 0, sizeof(answer));

	/*
	 *  If the name we're looking for has no spaces, treat it
	 *  as a surname, not a commonName.
	 */
	for (cp = who; *cp != '\0'; cp++) {
		if (isspace(*cp)) {
			spaces = 1;
			break;
		}
	}
	
	if (verbose) {
		register char *rcp = who;
		int seperator = 0;
		static char verbuf[8];

		for ( ; *rcp != '\0'; rcp++) {
			if ((*rcp == '.') || (*rcp == '_') || (*rcp == '"')) {
				seperator = 1;
				break;
			}
		}
		if (seperator) {
			printf("  There is no need to use periods, underscores, or quotes.\n");
			printf("  Shall I remove them for you? [Y/N] ");
			fflush(stdout);
			fetch_buffer(verbuf, sizeof(verbuf), stdin);
			if ((verbuf[0] == 'y') || (verbuf[0] == 'Y')) {
				for (rcp = who; *rcp != '\0'; rcp++) {
					if ((*rcp == '.') || (*rcp == '_') || (*rcp == '"')) {
						*rcp = ' ';
						/*
						 *  Aha!  There were spaces
						 *  after all.
						 */
						spaces = 1;
					}
				}
			}
		}
	}
			
again:
	/*
	 *  At this point, we could be trying a name for the first time, or
	 *  we could be retrying a name that has failed once already.  If
	 *  we haven't failed (i.e., try_fuzzy is 0), we can approach the
	 *  name as a surname if it has no spaces, or we can treat it as
	 *  a common name if there were spaces.
	 */
	if (!try_fuzzy) {
		if (spaces)
			sprintf(filter, "cn=%s", who);
		else
			sprintf(filter, "sn=%s", who);
	}
	else {
		if (spaces) {
			char *surname;
			surname = strchr(who, ' ');
			if (surname == NULL)
				sprintf(filter, "cn~=%s", who);
			else
				sprintf(filter, "cn~=%s | sn=%s", who, surname);
		}
		else
			sprintf(filter, "sn~=%s", who);
	}

	/*
	 *  Now perform the search and count the results.
	 */
	if (dxi_search(base, DXI_SRCH_SUBTREE, filter, attrs, answer) < 0) {
		dxi_perror("  dxi_search");
		return(NULL);
	}
	if ((matches = dxi_matches(answer)) < 0) {
		dxi_perror("  dxi_matches");
		fprintf(stderr, "  Buffer contained:\n");
		dxi_bprint(answer);
		return(NULL);
	}
	else if (matches == 1) {
		if (dxi_read(dxi_getdn(dxi_firstentry(answer)), read_attrs, answer) < 0) {
			dxi_perror("  dxi_read");
			return(NULL);
		}
		return(answer);
	}
	else if (matches == 0) { /* that didn't work -- try an inexact match */
		/*
		 *  Boo hoo!  No matches.  If we already tried a fuzzy match,
		 *  then we fall through.
		 */
		if (!try_fuzzy) {
			try_fuzzy = 1;
			goto again;
		}
		/*
		 *  If we're here, we're finished unless . . .
		 *
		 *  If we have been trying the name as a surname, and then
		 *  as a fuzzy surname, we still might match it as a common
		 *  name (or a fuzzy common name).
		 */
		if (!strncmp(filter, "sn~=", 4) && !spaces) {
			spaces = 1;		/* treat as a common name */
			try_fuzzy = 0;
			goto again;
		}
		printf(" Could not find \"%s\".\n", who);
		return(NULL);
	}

	/*
	 *  If we are here, it means that we got back multiple answers.
	 *  This can happen with both fuzzy and exact matches.
	 */
	if (dxi_errno == DXI_LIMIT) {
		if (verbose) {
			printf("  Your query was too general and a limit was exceeded.  The results listed\n");
			printf("  are not complete.  You may want to try again with a more refined query.\n\n");
		}
		else
			printf("  Time or size limit exceeded.  Partial results follow.\n\n");
	}
	printf("  %1d names matched \"%s\".\n", matches, who);
	for (;;) {
		printf("  Do you wish to see a list of names? ");
		fflush(stdout);
		(void) memset(response, 0, sizeof(response));
		fetch_buffer(response, sizeof(response), stdin);
		switch (response[0]) {
		case 'n' :
		case 'N' :
		case '\0' :
		case '\n' :
			return(NULL);
			/* NOTREACHED */
		case 'y' :
		case 'Y' :
			print_list(answer, namelist, &matches);
			if (num_picked == 0)
				choice = pick_one(matches);
			else
				choice = num_picked;
			num_picked = 0;
			if (choice < 0)
				return(NULL);
			if (dxi_read(namelist[choice], read_attrs, answer) < 0) {
				dxi_perror("dxi_read");
				return(NULL);
			}
			return(answer);
			/* NOTREACHED */
		default :
			printf("  What?\n");
			break;
		}
	}
	/* NOTREACHED */
}

pick_one(i)
int i;
{
	int n;
	char user_pick[RESP_SIZE];

	/* make the user pick an entry */
	for (;;) {
		printf("  Enter the number of the name you want or Q to quit: ");
		fflush(stdout);
		fetch_buffer(user_pick, sizeof(user_pick), stdin);
		if (user_pick[0] == 'q' || user_pick[0] == 'Q')
			return(-1);
		n = atoi(user_pick);
		if ((n > 0) && (n <= i))
			return(n);
		printf("  Invalid response\n");
	}
	/* NOTREACHED */
}

print_list(list, names, matches)
char *list, *names[];
int *matches;
{
	extern int lpp;
	char resp[RESP_SIZE];
	register char *ep, *s;
	register int i = 1;
	register int rest = 4;		/* 4, not 1 */

	/* print a list of names from which the user will select */
	for (ep = dxi_firstentry(list); ep != NULL; ep = dxi_nextentry()) {
		names[i] = dxi_getdn(ep);
		s = strrchr(dxi_getdn(ep), '=') + 1;
		printf(" %3d. %s\n", i, s);
		i++;
		if ((rest++ > (lpp - 1)) && (i < *matches)) {
again:
			printf("  More? ");
			fflush(stdout);
			fetch_buffer(resp, sizeof(resp), stdin);
			if ((resp[0] == 'n') || (resp[0] == 'N'))
				break;
			else if ((num_picked = atoi(resp)) != 0) {
				if (num_picked < i)
					break;
				else
					goto again;
			}
			rest = 1;
		}
	}
	*matches = i - 1;
	names[i] = NULL;
	return;
}

#ifdef UOFM

char * fetch_update_setting(who)
char *who;
{
	int setting;			/* 1 if TRUE, 0 if FALSE */
	char result[128];		/* should be plenty big */
	register char *ep, *ap, **vp;	/* for parsing answer */
	static char *attributes[] = { UPDATES, NULL };

	if (dxi_read(who, attributes, result) < 0) {
		if (dxi_errno == DXI_NEX_ATTR)
			return("FALSE");
		dxi_perror("dxi_read");
		return(NULL);
	}

	/*
	 *  We did a read on one name and only asked for one attribute.
	 *  There's no reason to loop through any of these structures.
	 *
	 *  We also don't handle the case where noBatchUpdates has more
	 *  than one value.  That should never happen anyhow.
	 */
	ep = dxi_firstentry(result);
	ap = dxi_firstattr(ep);
	if (!strcasecmp(ap, UPDATES)) {
		vp = (char **) dxi_getvalues(ep, ap);
		if (!strcasecmp(*vp, "TRUE"))
			setting = 1;
		else if (!strcasecmp(*vp, "FALSE"))
			setting = 0;
		else {
			fprintf(stderr, "  Got garbage -> [%s]\n", *vp);
			dxi_valuefree(vp);
			return(NULL);
		}
		dxi_valuefree(vp);
		if (setting == 1)
			return("TRUE");
		else
			return("FALSE");
	}
	else {
		fprintf(stderr, "  Got garbage -> [%s]\n", ap);
		return(NULL);
	}
}

#endif
