/*
 * Copyright (c) 1991 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include <string.h>
#ifndef DOS
#include <sgtty.h>
#endif
#include <ctype.h>
#include <signal.h>
#include <setjmp.h>
#if defined(NeXT)
#include <stdlib.h>
#else
#include <malloc.h>
#endif
#if defined(sparc) || defined(NeXT)
#include <stdlib.h>
#else
extern char * getenv();
#endif

#include <dixie.h>
#include "ud.h"

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1991 Regents of the University of Michigan.\nAll rights reserved.\n";
#endif

static char server[BUFSIZ]; 	/* dixie server name */

char default_base[BUFSIZ];	/* default search base set at start up */
extern char base[];		/* search base */
extern char me[];		/* string containing result of I AM */
static jmp_buf env;		/* spot to jump to on an interrupt */
int lpp;			/* lines per page */
int verbose;			/* 1 if verbose mode on */

main(argc, argv)
int argc;
char *argv[];
{
	extern char Version[];			/* version number */
	extern char *optarg;			/* for parsing argv */
	register int c;				/* for parsing argv */
	register char *cp;			/* for parsing Version */

	/*  handle argument list */
	while ((c = getopt(argc, argv, "d:s:v")) != -1)
		switch (c) {
		case 'd' :
			dxi_debug = atoi(optarg);
			break;
		case 's' :
			strcpy(server, optarg);
			break;
		case 'v' :
			verbose = 1;
			break;
		default:
			fprintf(stderr, "Usage: %s [-d debug-level] [-s server] [-v]\n",
					 argv[0]);
			exit(-1);
			/* NOTREACHED */
		}
	initialize_client();

	/* just print the first line of Version[] */
	cp = strchr(Version, '\t');
	if (cp != NULL)
		*cp = '\0';
	printf(Version);

	/* now tackle the user's commands */
	do_commands();
	/* NOTREACHED */
}

do_commands()
{
	register char *cp;			/* misc char pointer */
	register char *ap;			/* misc char pointer */
	static char input[BUFSIZ];		/* buffer for fgets() */
	extern struct entry Person;		/* holds info about someone */
	extern char *find(), *alias();

	if (verbose) 
		printf("\n  Enter a command.  If you need help, type 'h' or '?' and hit RETURN.\n");
	/* jump here on an interrupt */
	(void) setjmp(env);
	for (;;) {
		printf("* ");
		fflush(stdout);
		cp = input;
		fetch_buffer(input, sizeof(input), stdin);
		if (*input == '\0') {
			putchar('\n');
			continue;
		}
		while (isspace(*cp))
			cp++;	
		if (!strncasecmp(cp, "st", 2) && strncasecmp(cp, "stop", 4)) {
			status();
		}
		else if (*cp == 'c') {
			while (!isspace(*cp) && (*cp != '\0')) cp++;
			while (isspace(*cp) && (*cp != '\0')) cp++;
			if (!strncasecmp(cp, "base", 4))
				cp += 4;
			while (isspace(*cp) && (*cp != '\0')) cp++;
			if (*cp == '\0') {
				printf("  Syntax error\n");
				continue;
			}
			change_base(cp);
		}
		else if ((*cp == 'q') || !strcasecmp(cp, "stop")) {
			break;
		}
		else if (*cp == 'f') {
			cp++;
			while (!isspace(*cp) && (*cp != '\0'))
				cp++;
			while (isspace(*cp))
				cp++;
			if (*cp == '\0') {
				printf("  Syntax error\n");
				continue;
			}
			if ((cp = alias(cp)) == NULL)
				continue;
			if ((ap = find(cp)) != NULL) {
				parse_answer(ap);
				print_person();
			}
		}
		else if (*cp == 'd') {
			cp++;
			while (!isspace(*cp) && (*cp != '\0'))
				cp++;
			while (isspace(*cp))
				cp++;
			if (*cp == '\0') {
				printf("  Syntax error\n");
				continue;
			}
			if ((cp = alias(cp)) == NULL)
				continue;
			if ((ap = find(cp)) != NULL) {
				parse_answer(ap);
				print_person();
			}
		}
		else if (*cp == 's') {
			cp++;
			while (!isspace(*cp) && (*cp != '\0'))
				cp++;
			while (isspace(*cp))
				cp++;
			if (*cp == '\0') {
				printf("  Syntax error\n");
				continue;
			}
			if ((cp = alias(cp)) == NULL)
				continue;
			if ((ap = find(cp)) != NULL) {
				parse_answer(ap);
				print_person();
			}
		}
		else if (*cp == 'm') {
			cp++;
			while (!isspace(*cp) && (*cp != '\0'))
				cp++;
			while (isspace(*cp))
				cp++;
			(void) modify(alias(cp));
		}
		else if (*cp == 'b') {
			cp++;
			while (!isspace(*cp) && (*cp != '\0'))
				cp++;
			while (isspace(*cp))
				cp++;
			(void) auth(alias(cp), (char *) NULL);
		}
		else if ((*cp == '?') || (*cp == 'h')) {
			while (!isspace(*cp) && (*cp != '\0'))
				cp++;
			while (isspace(*cp))
				cp++;
			print_help(cp);
		}
		else if (*cp == 'i') {
			while (!isspace(*cp) && (*cp != '\0'))
				cp++;
			while (isspace(*cp))
				cp++;
			iam(cp);
		}
		else if (*cp == 'w') {
			print_my_name();
		}
		else if (*cp == 'v') {
			verbose = 1 - verbose;
			/*  ;-)  */
			if (verbose)
				printf("  Verbose mode has been turned on.\n");
		}
		else if (*cp == '\0') {
			putchar('\n');
		}
		else {
			printf("  Invalid command.  Type \"help commands.\"\n");
		}
	}
	printf(" Thank you!\n");
	dxi_close();
	exit(0);
	/* NOTREACHED */
}

status()
{
	void printbase();

	printf("  Current server is %s\n", server);
	printbase("  Search base is ", base);
	bind_info();
}

change_base(s)
char *s;
{
	int num_picked;				/* # of selected base */
	int j;					/* used with num_picked */
	int i = 1;				/* index into choices array */
	int matches;				/* # of matches found */
	int rest = 1;				/* # left to display */
	register char *ep;			/* for stepping thru bases */
	extern char answer[MAX_PKT_SIZE];	/* returned from dxi_list() */
	static char *choices[BUFSIZ];		/* bases from which to choose */
	static char resp[16];			/* for prompting user */
	void printbase();

	/* sanity check -- should never happen */
	if (s == NULL) {
		fprintf(stderr, "NULL base!\n");
		dxi_close();
		exit(0);
	}

	/* user forgot to supply an argument on the change base command */
	if ((*s == '\n') || (*s == '\0'))
		return;

	/* specifying 'root' for '@' is OK */
	if (!strcasecmp(s, "root")) {
		base[0] = '@';
		base[1] = '\0';
		printbase("  Search base is ", base);
		return;
	}

	/* user wants to move up one level -- easy */
	if (!strcasecmp(s, "..")) {
		char *cp;

		if ((cp = strrchr(base, '@')) == NULL) {
			base[0] = '@';
			base[1] = '\0';
		}
		else
			*cp = '\0';
		printbase("  Search base is ", base);
		return;
	}

	/* user wants to see what is directly below this level */
	if (*s == '?') {
		/*
		 *  Fetch the list of entries directly below this level.
		 *  Once we have the list, we will print it for the user, one
		 *  screenful at a time.  At the end of each screen, we ask
		 *  the user if they want to see more.  They can also just
		 *  type a number at that point too.
		 */
		if (dxi_list(base, answer) < 0) {
			if (dxi_errno == DXI_REFER_ERROR) {
				printbase("  The X.500 server for ", base);
				printf("  is unavailable at this time.  Try again later.\n");
				if (verbose)
					printbase("  Search base is ", base);
			}
			else
				dxi_perror("dxi_list");
			return;
		}
		matches = dxi_matches(answer);
		num_picked = 0;
		printf(" There are %d choices:\n", matches);
		for (ep = dxi_firstentry(answer); ep != NULL; ep = dxi_nextentry()) {
			choices[i] = dxi_getdn(ep);
			printf(" %2d. %s\n", i, dxi_lastdncomp(choices[i]));
			i++;
			if ((rest++ > (lpp - 3)) && (i < matches)) {
				printf("More? ");
				fflush(stdout);
				fetch_buffer(resp, sizeof(resp), stdin);
				if ((resp[0] == 'n') || (resp[0] == 'N'))
					break;
				else if (((num_picked = atoi(resp)) != 0) && (num_picked < i))
					break;
				rest = 1;
			}
		}
		for (;;) {
			if (num_picked != 0) {
				j = num_picked;
				num_picked = 0;
			}
			else {
				printf(" Which number? ");
				fflush(stdout);
				fetch_buffer(resp, sizeof(resp), stdin);
				j = atoi(resp);
			}
			if (j == 0)
				return;
			if ((j < 1) || (j >= i))
				printf(" Invalid number\n");
			else {
				sprintf(base, "%s@%s", base, choices[j]);
				printbase("  Search base is ", base);
				return;
			}
		}
	}
	/* set the search base back to the original default value */
	else if (!strcasecmp(s, "default")) {
		strcpy(base, default_base);
		printf("  Search base is %s\n", base);
	}
	/* they typed in something -- see if it is legit */
	else {
		/* user cannot do something like 'cb 33' */
		if (atoi(s) != 0) {
			printf("  \"%s\" is not a valid search base\n", s);
			printf("  Search base unchanged.\n");
			printf("  Try using 'cb ?'\n");
			return;
		}
		if (dxi_vrfy(s)) {
			strcpy(base, s);
			printbase("  Search base is ", base);
		}
		else {
			printf("  \"%s\" is not a valid search base\n", s);
			printf("  Search base unchanged.\n");
		}
	}
}

initialize_client()
{
	FILE *fp;				/* for config file */
	static char buffer[BUFSIZ];		/* for fgets() */
	register char *cp;			/* for fiddling with fgets() */
	char *term;				/* for tty set-up */
	static char bp[1024];			/* for tty set-up */
	extern char bound_as[];			/* who we are bound as */
	extern SIG_FN attn();			/* ^C signal handler */

	/*
	 *  If there is a config file, read it.
	 *
	 *  Could have lines that look like this:
	 *
	 *	server <ip-address or domain-name>
	 *	base   <default search base>
	 */
	strcpy(default_base, DEFAULT_BASE);
	if ((fp = fopen(CONFIG_FILE, "r")) != NULL) {
		while (fgets(buffer, sizeof(buffer), fp) != NULL) {
			if (!strncmp(buffer, "server", 6)) {
				if (server[0] == '\0') {
					cp = buffer + 6;
					while (isspace(*cp))
						cp++;
					if ((*cp == '\0') || (*cp == '\n'))
						continue;
					(void) strncpy(server, cp, sizeof(server) - 1);
					if ((cp = strchr(server, '\n')) != NULL)
						*cp = '\0';
				}
			}
			else if (!strncmp(buffer, "base", 4)) {
				cp = buffer + 4;
				while (isspace(*cp))
					cp++;
				if ((*cp == '\0') || (*cp == '\n'))
					continue;
				(void) strncpy(default_base, cp, sizeof(default_base) - 1);
				if ((cp = strchr(default_base, '\n')) != NULL)
					*cp = '\0';
			}
			else
				fprintf(stderr, "?? -> %s\n", buffer);
		}
	}
	if (server[0] == '\0')
		strcpy(server, DEFAULT_SERVER);

	/*
	 *  Set up our DIXIE connection.  The values of retry and timeout
	 *  are meaningless since we will immediately be doing a null bind
	 *  because we want to be sure to use TCP, not UDP.
	 */
	if (dxi_init(server, 3, 5) < 0) {
		fprintf(stderr, "  Could not contact '%s'.\n", server);
		dxi_close();
		exit(0);
		/* NOTREACHED */
	}
	dxi_bind((char *) NULL, (char *) NULL);
	(void) strcpy(bound_as, "Nobody");

	/* set the initial search base */
	strcpy(base, default_base);

	/* terminal initialization stuff goes here */
	lpp = DEFAULT_LINES;

	(void) signal(SIGINT, attn);

#if !defined(DOS) && !defined(NOTERMCAP)
	{
	struct winsize win;			/* for tty set-up */
	extern SIG_FN chwinsz();		/* WINSZ signal handler */

	if (((term = getenv("TERM")) == NULL) || (tgetent(bp, term) <= 0))
		return;
	else {
		if (ioctl(fileno(stdout), TIOCGWINSZ, &win) < 0) {
			lpp = tgetnum("li");
		}
		else {
			if ((lpp = win.ws_row) == 0)
				lpp = tgetnum("li");
			if ((lpp <= 0) || tgetflag("hc"))
				lpp = DEFAULT_LINES;
		}
	}
	(void) signal(SIGWINCH, chwinsz);

	}
#endif
}

SIG_FN attn()
{
	fflush(stderr);
	fflush(stdout);
	printf("\n\n  INTERRUPTED!\n");
#if defined(DOS) || defined(SYSV)
	(void) signal(SIGINT, attn);
#endif
	longjmp(env, 1);
}

print_help(s)
char *s;
{
	/* print general help, or just on topic 's' if provided */
	if (*s == '\0') {
		printf("\n  Here are brief descriptions of available commands:\n\n");
		printf("  ?             To print this list.\n");
		printf("  bind          To bind (authenticate) to the Directory.\n");
		printf("  change base   To change the search base.\n");
		printf("  display       To display information from the Directory.\n");
		printf("  help          To print this list.\n");
		printf("  iam           To set your name while running the program.\n");
		printf("  find          To find an entry in the Directory\n");
		printf("  modify        To change information associated with an entry.\n");
		printf("  quit          To terminate the program.\n");
		printf("  stop          To terminate the program.\n");
		printf("  verbose       To toggle the verbose switch.\n");
		printf("  whoami        To print who you are (used with 'iam').\n");

		printf("\n  Type \"help <command-name>\" to get help about a particular command.\n\n");
	}
	else {
		switch(*s) {
		case 's' :
			printf("  status\n\n");
printf("  Prints out the current connection status.\n");
printf("  Lists the name of the current DIXIE server, the current search\n");
printf("  base, and who you are bound as.  If you have not bound as anyone\n");
printf("  then ud considers you bound as Nobody.\n\n");
			break;
		case 'c' :
			printf("\n  change base <to-what>\n\n");
printf("  The CHANGE BASE command changes the search base.  By default, this\n");
printf("  program looks only in the local part of the Directory.  By using the\n");
printf("  CHANGE BASE command, you can search other parts of the Directory.\n\n");
printf("  Examples:\n\n");
printf("            * change base ..\n\n");
printf("  This example changes the search base so that it is one level higher\n");
printf("  in the Directory.  Note that if you perform several of these in a row\n");
printf("  you will move to the root of the Directory tree.\n\n");
printf("            * change base ?\n\n");
printf("  This example prints out a list of the possible areas to search below\n");
printf("  the current search base.  This is useful once you have moved high in the\n");
printf("  tree and wish to snoop about.\n\n");
printf("            * change base default\n\n");
printf("  This example sets the search base to its original default value.\n\n");
printf("            * change base <new-search-base>\n\n");
printf("  This example sets the search base to the value specified.  This program\n");
printf("  inspects the specified search base for validity, and if it is not valid\n");
printf("  it does not set the search base.  The search base must be a valid\n");
printf("  Distinguished Name.  If you do not know what a DN is, do not use this\n");
printf("  form of the command.\n\n");
			break;
		case 'q' :
			printf("  quit ud\n\n");
printf("  Quits the program.\n");
			break;
		case 'd' :
		case 'f' :
			printf("  display information about someone\n\n");
printf("  Displays information about the person specified.  If 'ud' matches\n");
printf("  more than one person, it will give you a list from which to make\n");
printf("  your choice.\n\n");
			break;
		case 'b' :
			printf("\n  bind [who]\n\n");
printf("  The BIND command is used to bind (authenticate) to the Directory.  It\n");
printf("  is generally necessary to bind to the Directory in order to look at\n");
printf("  privileged entries or to modify an entry.\n\n");
printf("  If you have not used another command like \"find\" or \"display\"\n");
printf("  or \"show\" to look up someone in the Directory, you will need to\n");
printf("  specify a \"who\" with the BIND command.  This tells the program\n") ;
printf("  to whom you wish to bind.  If you have already used a command to\n");
printf("  display some information from the Directory, the program will make\n");
printf("  an attempt to specify a \"who\" for you.\n\n");
printf("  Examples:\n\n");
printf("            * bind\n");
printf("            Authenticate as [????]?\n\n");
printf("  In this example, no \"who\" was specified, and the user had not already\n");
printf("  looked someone up in the Directory.  At this point you should enter the\n");
printf("  name to which you wish to bind.\n\n");
printf("            * bind\n");
printf("            Authenticate as [John Doe]?\n\n");
printf("  In this example, the user had already looked up John Doe.  If the user\n");
printf("  wishes to bind as John Doe, s/he should simply hit RETURN.  Otherwise,\n");
printf("  the user should enter the name to which s/he wishes to bind.\n");

			break;
		case 'm' :
			printf("\n  modify [who]\n\n");
printf("  The MODIFY command changes information associated with an entry\n");
printf("  in the X.500 Directory.\n\n");
printf("  If you have not used another command like \"find\" or \"display\"\n");
printf("  or \"show\" to look up someone in the Directory, you will need to\n");
printf("  specify a \"who\" with the MODIFY command.  This tells the program\n");
printf("  whose information you wish to modify.  If you have already used\n");
printf("  a command to display some information from the Directory, the program\n");
printf("  will make an attempt to specify a \"who\" for you.\n\n");
printf("  In either case, once you have typed the \"modify\" command, the program\n");
printf("  will prompt you for a field to modify like this:\n\n");
printf("          * modify Bryan Beecher\n");
printf("            Modify what?\n\n");
printf("  At this point, hit a RETURN to cancel or a \"?\" to see a list of\n");
printf("  possible attributes (e.g., Business Address) to modify.\n\n");
printf("  See the BIND command for more information about MODIFY.\n");
			break;
		case 'i' :
			printf(" set your name\n\n");
printf("  Used to set your name so 'ud' knows who you are.  You may then\n");
printf("  use 'me' as a legal identifier when issuing other commands.\n");
			break;
		case 'w' :
			printf(" print your name\n\n");
printf("  Used to print the name you set with the 'iam' command.\n");
			break;
		case 'v' :
			printf(" print verbose messages\n\n");
printf("  Used to turn on long and windy messages which might be useful to\n");
printf("  new users of this program.  If verbose mode is already on, this turns it off.\n");
			break;
		case 'h' :
		case '?' :
printf("  Used to print out helpful information.\n");
			break;
		default :
			printf("Don't recognize <%s>\n", s);
			break;
		}
	}
}

#if !defined(DOS) && !defined(NOTERMCAP)
SIG_FN chwinsz() 
{
	struct winsize win;

	(void) signal(SIGWINCH, SIG_IGN);
	if (ioctl(fileno(stdout), TIOCGWINSZ, &win) != -1) {
		if (win.ws_row != 0)
			lpp = win.ws_row;
	}
	(void) signal(SIGWINCH, chwinsz);
}
#endif

char * alias(s)
char *s;
{
	if (!strncasecmp("me", s, 2) && (*(s + 2) == '\0')) {
		if (me[0] == '\0') {
			printf("  I don't know who you are.\n");
			return(NULL);
		}
		else
			return(me);
	}
	else if (*s == '\0')
		return(NULL);
	else
		return(s);
}


void printbase(lead, s)
char *lead, *s;
{
	char *work;
	register char *cp;

	/* handle easy case first */
	if (*s == '@') {
		printf("%sroot\n", lead);
		return;
	}

	/* make a copy of the search base */
	if ((work = (char *) malloc(strlen(s) + 1)) == NULL) {
		perror("malloc");
		printf("<cannot print base>\n");
		return;
	}
	strcpy(work, s);
	
	/* now parse thru it, printing out bits at a time */
	printf(lead);
	for (cp = strrchr(work, '@'); cp != NULL; cp = strrchr(work, '@')) {
		*cp = '\0';
		while (*cp != '=')
			cp++;
		cp++;
		printf("%s, ", cp);
	}
	for (cp = work; *cp != '='; cp++)
		;
	cp++;
	printf("%s\n", cp);
	free(work);
	return;
}

fetch_buffer(buffer, length, where)
char *buffer;
int length;
FILE *where;
{
	register int i;

	/*
	 *  Fetch a buffer and strip off any trailing non-printing
	 *  characters, namely newlines and carriage returns.
	 */
	if (fgets(buffer, length, where) == NULL) {
		printf("\n Thank you!\n");
		dxi_close();
		exit(0);
	}
	for (i = strlen(buffer) - 1; i >= 0 && !isprint(buffer[i]); i--)
		buffer[i] = '\0';
}
